@@ -1,5 +1,5 @@
-/* 
- * Copyright (C) 2006, 2007 OpenedHand Ltd.
+/*
+ * Copyright (C) 2006, 2007, 2008 OpenedHand Ltd.
  *
  * Author: Jorn Baayen <jorn@openedhand.com>
  *
@@ -25,6 +25,9 @@
  *
  * #GUPnPContext wraps the networking bits that are used by the various
  * GUPnP classes. It automatically starts a web server on demand.
+ *
+ * For debugging, it is possible to see the messages being sent and received by
+ * exporting #GUPNP_DEBUG.
  */
 
 #include <config.h>
@@ -35,22 +38,21 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <sys/utsname.h>
-#include <sys/ioctl.h> 
+#include <sys/ioctl.h>
 #include <sys/socket.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <net/if.h>
-#include <netinet/in.h>
 #include <arpa/inet.h>
+#include <net/if.h>
+#include <ifaddrs.h>
 #include <libsoup/soup-address.h>
-
-#include "xdgmime/xdgmime.h"
+#include <glib/gstdio.h>
 
 #include "gupnp-context.h"
 #include "gupnp-context-private.h"
 #include "gupnp-marshal.h"
 #include "gena-protocol.h"
-#include "accept-language.h"
+#include "http-headers.h"
 
 G_DEFINE_TYPE (GUPnPContext,
                gupnp_context,
@@ -72,15 +74,16 @@ enum {
         PROP_0,
         PROP_HOST_IP,
         PROP_PORT,
+        PROP_SERVER,
         PROP_SUBSCRIPTION_TIMEOUT
 };
 
 #define LOOPBACK_IP "127.0.0.1"
 
-typedef char ip_address[15+1];
-
-#define inaddrr(x) (*(struct in_addr *) &ifr->x[sizeof sa.sin_port])
-#define IFRSIZE   ((int)(size * sizeof (struct ifreq)))
+typedef struct {
+        char *local_path;
+        char *server_path;
+} PathData;
 
 /**
  * Generates the default server ID.
@@ -91,84 +94,68 @@ make_server_id (void)
         struct utsname sysinfo;
 
         uname (&sysinfo);
-        
+
         return g_strdup_printf ("%s/%s UPnP/1.0 GUPnP/%s",
                                 sysinfo.sysname,
                                 sysinfo.version,
                                 VERSION);
 }
 
-/**
- * Looks up the IP address for interface @iface.
- *
- * http://www.linuxquestions.org/questions/showthread.php?t=425637
- **/
-gboolean
-get_ip (const char *iface, ip_address ip)
-{
-        struct ifreq *ifr;
-        struct ifreq ifrr;
-        struct sockaddr_in sa;
-        struct sockaddr ifaddr;
-        int sockfd;
-
-        if ((sockfd = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1)
-                return FALSE;
-
-        ifr = &ifrr;
-
-        ifrr.ifr_addr.sa_family = AF_INET;
-
-        strncpy (ifrr.ifr_name, iface, sizeof (ifrr.ifr_name));
-
-        if (ioctl (sockfd, SIOCGIFADDR, ifr) < 0)
-                return FALSE;
-
-        ifaddr = ifrr.ifr_addr;
-        strncpy (ip, inet_ntoa (inaddrr (ifr_addr.sa_data)),
-                 sizeof (ip_address));
-
-        return TRUE;
-}
-
 /**
  * Looks up the IP address of the first non-loopback network interface.
  **/
 static char *
 get_default_host_ip (void)
 {
-        struct if_nameindex *ifs;
+        struct ifaddrs *ifa_list, *ifa;
         char *ret;
-        int i;
-        
-        /* List network interfaces */
-        ifs = if_nameindex ();
-        if (ifs == NULL) {
+
+        ret = NULL;
+
+        if (getifaddrs (&ifa_list) != 0) {
                 g_error ("Failed to retrieve list of network interfaces:\n%s\n",
                          strerror (errno));
 
                 return NULL;
         }
 
-        ret = NULL;
+        for (ifa = ifa_list; ifa != NULL; ifa = ifa->ifa_next) {
+                char ip[INET6_ADDRSTRLEN];
+                const char *p;
+                struct sockaddr_in *s4;
+                struct sockaddr_in6 *s6;
 
-        for (i = 0; ifs[i].if_index != 0 && ifs[i].if_name != NULL; i++) {
-                ip_address ip;
-
-                /* Retrieve IP address */
-                if (!get_ip (ifs[i].if_name, ip))
+                if (ifa->ifa_addr == NULL)
                         continue;
 
-                if (strcmp (ip, LOOPBACK_IP) != 0) {
-                        /* This is not a loopback interface - just what we
-                         * are looking for. */
-                        ret = g_strdup (ip);
+                if ((ifa->ifa_flags & IFF_LOOPBACK) ||
+                    !(ifa->ifa_flags & IFF_UP))
+                        continue;
 
+                p = NULL;
+
+                switch (ifa->ifa_addr->sa_family) {
+                case AF_INET:
+                        s4 = (struct sockaddr_in *) ifa->ifa_addr;
+                        p = inet_ntop (AF_INET,
+                                       &s4->sin_addr, ip, sizeof (ip));
+                        break;
+                case AF_INET6:
+                        s6 = (struct sockaddr_in6 *) ifa->ifa_addr;
+                        p = inet_ntop (AF_INET6,
+                                       &s6->sin6_addr, ip, sizeof (ip));
+                        break;
+                default:
+                        continue; /* Unknown: ignore */
+                }
+
+                if (p != NULL) {
+                        ret = g_strdup (p);
                         break;
                 }
         }
 
-        if_freenameindex (ifs);
+        freeifaddrs (ifa_list);
 
         if (!ret) {
                 /* Didn't find anything. Let's take the loopback IP. */
@@ -190,9 +177,20 @@ gupnp_context_init (GUPnPContext *context)
 
         context->priv->session = soup_session_async_new ();
 
+        g_object_set (context->priv->session,
+                      SOUP_SESSION_IDLE_TIMEOUT,
+                      60,
+                      NULL);
+
         server_id = make_server_id ();
         gssdp_client_set_server_id (GSSDP_CLIENT (context), server_id);
         g_free (server_id);
+
+	if (g_getenv ("GUPNP_DEBUG")) {
+		SoupLogger *logger;
+		logger = soup_logger_new (SOUP_LOGGER_LOG_BODY, -1);
+		soup_logger_attach (logger, context->priv->session);
+	}
 }
 
 static void
@@ -240,6 +238,10 @@ gupnp_context_get_property (GObject    *object,
                 g_value_set_uint (value,
                                   gupnp_context_get_port (context));
                 break;
+        case PROP_SERVER:
+                g_value_set_object (value,
+                                    gupnp_context_get_server (context));
+                break;
         case PROP_SUBSCRIPTION_TIMEOUT:
                 g_value_set_uint (value,
                                   gupnp_context_get_subscription_timeout
@@ -341,11 +343,29 @@ gupnp_context_class_init (GUPnPContextClass *klass)
                                     G_PARAM_STATIC_NICK |
                                     G_PARAM_STATIC_BLURB));
 
+        /**
+         * GUPnPContext:server
+         *
+         * The #SoupServer HTTP server used by GUPnP.
+         **/
+        g_object_class_install_property
+                (object_class,
+                 PROP_SERVER,
+                 g_param_spec_object ("server",
+                                      "SoupServer",
+                                      "SoupServer HTTP server",
+                                      SOUP_TYPE_SERVER,
+                                      G_PARAM_READABLE |
+                                      G_PARAM_STATIC_NAME |
+                                      G_PARAM_STATIC_NICK |
+                                      G_PARAM_STATIC_BLURB));
+
         /**
          * GUPnPContext:subscription-timeout
          *
          * The preferred subscription timeout: the number of seconds after
-         * which subscriptions are renewed.
+         * which subscriptions are renewed. Set to '0' if subscriptions 
+         * are never to time out.
          **/
         g_object_class_install_property
                 (object_class,
@@ -353,8 +373,8 @@ gupnp_context_class_init (GUPnPContextClass *klass)
                  g_param_spec_uint ("subscription-timeout",
                                     "Subscription timeout",
                                     "Subscription timeout",
-                                    GENA_MIN_TIMEOUT,
-                                    G_MAXUINT,
+                                    0,
+                                    GENA_MAX_TIMEOUT,
                                     GENA_DEFAULT_TIMEOUT,
                                     G_PARAM_READWRITE |
                                     G_PARAM_CONSTRUCT_ONLY |
@@ -373,27 +393,41 @@ _gupnp_context_get_session (GUPnPContext *context)
  * Default server handler: Return 404 not found.
  **/
 static void
-default_server_handler (SoupServerContext *server_context,
-                        SoupMessage       *msg,
+default_server_handler (SoupServer        *server,
+                        SoupMessage       *msg, 
+                        const char        *path,
+                        GHashTable        *query,
+                        SoupClientContext *client,
                         gpointer           user_data)
 {
         soup_message_set_status (msg, SOUP_STATUS_NOT_FOUND);
 }
 
+/**
+ * gupnp_context_get_server
+ * @context: A #GUPnPContext
+ *
+ * Get the #SoupServer HTTP server that GUPnP is using.
+ *
+ * Return value: The #SoupServer used by GUPnP.  Do not unref this when finished.
+ **/
 SoupServer *
-_gupnp_context_get_server (GUPnPContext *context)
+gupnp_context_get_server (GUPnPContext *context)
 {
+        g_return_val_if_fail (GUPNP_IS_CONTEXT (context), NULL);
+
         if (context->priv->server == NULL) {
                 context->priv->server = soup_server_new (SOUP_SERVER_PORT,
                                                          context->priv->port,
                                                          NULL);
 
-                soup_server_add_handler (context->priv->server, NULL, NULL,
-                                         default_server_handler, NULL, context);
+                soup_server_add_handler (context->priv->server, NULL,
+                                         default_server_handler, context,
+                                         NULL);
 
                 soup_server_run_async (context->priv->server);
         }
-        
+
         return context->priv->server;
 }
 
@@ -410,7 +444,7 @@ make_server_url (GUPnPContext *context)
                 context->priv->host_ip = get_default_host_ip ();
 
         /* What port are we running on? */
-        server = _gupnp_context_get_server (context);
+        server = gupnp_context_get_server (context);
         port = soup_server_get_port (server);
 
         /* Put it all together */
@@ -432,9 +466,11 @@ _gupnp_context_get_server_url (GUPnPContext *context)
  * @host_ip: The local host's IP address, or NULL to use the IP address
  * of the first non-loopback network interface.
  * @port: Port to run on, or 0 if you don't care what port is used.
- * @error: A location to store a #GError, or NULL
+ * @error: A location to store a #GError, or %NULL
  *
- * Return value: A new #GUPnPContext object.
+ * Create a new #GUPnPContext with the specified @main_context, @host_ip and @port.
+ *
+ * Return value: A new #GUPnPContext object, or %NULL on an error
  **/
 GUPnPContext *
 gupnp_context_new (GMainContext *main_context,
@@ -454,7 +490,9 @@ gupnp_context_new (GMainContext *main_context,
  * gupnp_context_get_host_ip
  * @context: A #GUPnPContext
  *
- * Return value: The IP address we advertise ourselves as using. Do not free.
+ * Get the IP address we advertise ourselves as using.
+ *
+ * Return value: The IP address. This string should not be freed.
  **/
 const char *
 gupnp_context_get_host_ip (GUPnPContext *context)
@@ -471,6 +509,8 @@ gupnp_context_get_host_ip (GUPnPContext *context)
  * gupnp_context_get_port
  * @context: A #GUPnPContext
  *
+ * Get the port that the SOAP server is running on.
+ 
  * Return value: The port the SOAP server is running on.
  **/
 guint
@@ -480,7 +520,7 @@ gupnp_context_get_port (GUPnPContext *context)
 
         g_return_val_if_fail (GUPNP_IS_CONTEXT (context), 0);
 
-        server = _gupnp_context_get_server (context);
+        server = gupnp_context_get_server (context);
         return soup_server_get_port (server);
 }
 
@@ -508,8 +548,9 @@ gupnp_context_set_subscription_timeout (GUPnPContext *context,
  * gupnp_context_get_subscription_timeout
  * @context: A #GUPnPContext
  *
- * Return value: The event subscription timeout in seconds, or 0, meaning
- * no timeout.
+ * Get the event subscription timeout (in seconds), or 0 meaning there is no timeout.
+ * 
+ * Return value: The event subscription timeout in seconds.
  **/
 guint
 gupnp_context_get_subscription_timeout (GUPnPContext *context)
@@ -521,23 +562,26 @@ gupnp_context_get_subscription_timeout (GUPnPContext *context)
 
 /* Modified from libsoup simple-httpd.c */
 static void
-hosting_server_handler (SoupServerContext *server_context,
-                        SoupMessage       *msg,
+hosting_server_handler (SoupServer        *server,
+                        SoupMessage       *msg, 
+                        const char        *path,
+                        GHashTable        *query,
+                        SoupClientContext *client,
                         gpointer           user_data)
 {
-        const char *local_path, *lang, *mime;
-        char *path, *path_to_open, *path_locale, *slash;
-        SoupMethodId method;
+        PathData *path_data;
+        const char *lang;
+        char *path_to_open, *path_locale, *slash, *content_type, *mime;
         struct stat st;
-        int fd, path_offset;
+        int path_offset;
         GList *locales;
+        GMappedFile *mapped_file;
+        GError *error;
 
-        local_path = (const char *) user_data;
+        path_data = (PathData *) user_data;
 
-        path = soup_uri_to_string (soup_message_get_uri (msg), TRUE);
-
-        method = soup_method_get_id (msg->method);
-        if (method != SOUP_METHOD_ID_GET && method != SOUP_METHOD_ID_HEAD) {
+        if (msg->method != SOUP_METHOD_GET &&
+            msg->method != SOUP_METHOD_HEAD) {
                 soup_message_set_status (msg, SOUP_STATUS_NOT_IMPLEMENTED);
                 locales = NULL;
                 goto DONE;
@@ -553,7 +597,7 @@ hosting_server_handler (SoupServerContext *server_context,
                 }
 
                 /* Skip the server path */
-                path_offset = strlen (server_context->handler->path);
+                path_offset = strlen (path_data->server_path);
         } else {
                 path = g_strdup ("");
 
@@ -572,10 +616,12 @@ hosting_server_handler (SoupServerContext *server_context,
         } else
                 path_locale = g_strdup (path + path_offset);
 
-        path_to_open = g_build_filename (local_path, path_locale, NULL);
+        path_to_open = g_build_filename (path_data->local_path,
+                                         path_locale,
+                                         NULL);
 
  AGAIN:
-        if (stat (path_to_open, &st) == -1) {
+        if (g_stat (path_to_open, &st) == -1) {
                 g_free (path_to_open);
                 if (errno == EPERM)
                         soup_message_set_status (msg, SOUP_STATUS_FORBIDDEN);
@@ -601,8 +647,8 @@ hosting_server_handler (SoupServerContext *server_context,
                         uri = soup_uri_to_string (soup_message_get_uri (msg),
                                                   FALSE);
                         redir_uri = g_strdup_printf ("%s/", uri);
-                        soup_message_add_header (msg->response_headers,
-                                                 "Location", redir_uri);
+                        soup_message_headers_append (msg->response_headers,
+                                                     "Location", redir_uri);
                         soup_message_set_status (msg,
                                                  SOUP_STATUS_MOVED_PERMANENTLY);
                         g_free (redir_uri);
@@ -612,7 +658,7 @@ hosting_server_handler (SoupServerContext *server_context,
                 }
 
                 g_free (path_to_open);
-                path_to_open = g_build_filename (local_path,
+                path_to_open = g_build_filename (path_data->local_path,
                                                  path_locale,
                                                  "index.html", NULL);
                 goto AGAIN;
@@ -620,30 +666,107 @@ hosting_server_handler (SoupServerContext *server_context,
 
         g_free (path_locale);
 
-        /* Read file */
-        fd = open (path_to_open, O_RDONLY);
+        /* Map file */
+        error = NULL;
+        mapped_file = g_mapped_file_new (path_to_open, FALSE, &error);
+
         g_free (path_to_open);
-        if (fd == -1) {
+
+        if (mapped_file == NULL) {
+                g_warning ("Unable to map file %s: %s",
+                           path_to_open, error->message);
+
+                g_error_free (error);
+
                 soup_message_set_status (msg,
                                          SOUP_STATUS_INTERNAL_SERVER_ERROR);
+
                 goto DONE;
         }
 
-        msg->response.owner = SOUP_BUFFER_SYSTEM_OWNED;
-        msg->response.length = st.st_size;
+        if (msg->method == SOUP_METHOD_GET) {
+                guint64 offset, length;
+                gboolean have_range;
+                SoupBuffer *buffer;
 
-        msg->response.body = g_malloc (msg->response.length);
-        read (fd, msg->response.body, msg->response.length);
+                /* Find out range */
+                have_range = FALSE;
 
-        close (fd);
+                offset = 0;
+                length = st.st_size;
+
+                if (!range_get (msg, &have_range, &offset, &length)) {
+                        soup_message_set_status
+                                (msg,
+                                 SOUP_STATUS_REQUESTED_RANGE_NOT_SATISFIABLE);
+
+                        goto DONE;
+                }
+
+                if (have_range && (length < 0                   ||
+                                   offset < 0                   ||
+                                   length > st.st_size - offset ||
+                                   offset >= st.st_size)) {
+                        soup_message_set_status
+                                (msg,
+                                 SOUP_STATUS_REQUESTED_RANGE_NOT_SATISFIABLE);
+
+                        goto DONE;
+                }
+
+                /* Add requested content */
+                buffer = soup_buffer_new_with_owner
+                             (g_mapped_file_get_contents (mapped_file) + offset,
+                              length,
+                              mapped_file,
+                              (GDestroyNotify) g_mapped_file_free);
+
+                soup_message_body_append_buffer (msg->response_body, buffer);
+
+                soup_buffer_free (buffer);
+
+                /* Set status */
+                if (have_range) {
+                        char *content_range;
+
+                        content_range = g_strdup_printf
+                                ("bytes %" G_GUINT64_FORMAT "-%"
+                                 G_GUINT64_FORMAT "/%" G_GUINT64_FORMAT,
+                                 offset,
+                                 offset + length,
+                                 st.st_size);
+
+                        soup_message_headers_append (msg->response_headers,
+                                                     "Content-Range",
+                                                     content_range);
+
+                        g_free (content_range);
+
+                        soup_message_set_status (msg,
+                                                 SOUP_STATUS_PARTIAL_CONTENT);
+                } else
+                        soup_message_set_status (msg, SOUP_STATUS_OK);
+
+        } else /* SOUP_METHOD_HEAD */
+                soup_message_set_status (msg, SOUP_STATUS_OK);
 
         /* Set Content-Type */
-        mime = xdg_mime_get_mime_type_for_data (msg->response.body,
-                                                msg->response.length);
+        content_type = g_content_type_guess
+                                (path_to_open,
+                                 (guchar*)
+                                 g_mapped_file_get_contents (mapped_file),
+                                 st.st_size,
+                                 NULL);
+        mime = g_content_type_get_mime_type (content_type);
+        if (mime == NULL)
+                mime = g_strdup ("application/octet-stream");
 
-        soup_message_add_header (msg->response_headers,
-                                 "Content-Type",
-                                 mime);
+        soup_message_headers_append (msg->response_headers,
+                                     "Content-Type",
+                                     mime);
+
+        g_free (mime);
+        g_free (content_type);
 
         /* Set Content-Language */
         if (locales) {
@@ -653,29 +776,44 @@ hosting_server_handler (SoupServerContext *server_context,
         } else
                 lang = "en";
 
-        soup_message_add_header (msg->response_headers,
-                                 "Content-Language",
-                                 lang);
+        soup_message_headers_append (msg->response_headers,
+                                     "Content-Language",
+                                     lang);
 
-        /* OK */
-        soup_message_set_status (msg, SOUP_STATUS_OK);
+        /* Set Accept-Ranges */
+        soup_message_headers_append (msg->response_headers,
+                                     "Accept-Ranges",
+                                     "bytes");
 
  DONE:
         /* Cleanup */
-        g_free (path);
-
         while (locales) {
                 g_free (locales->data);
                 locales = g_list_delete_link (locales, locales);
         }
 }
 
-static void
-unregister_hosting_server_handler (SoupServer        *server,
-                                   SoupServerHandler *handler,
-                                   gpointer           user_data)
+PathData *
+path_data_new (const char *local_path,
+               const char *server_path)
 {
-        g_free (user_data);
+        PathData *path_data;
+
+        path_data = g_slice_new (PathData);
+
+        path_data->local_path  = g_strdup (local_path);
+        path_data->server_path = g_strdup (server_path);
+
+        return path_data;
+}
+
+void
+path_data_free (PathData *path_data)
+{
+        g_free (path_data->local_path);
+        g_free (path_data->server_path);
+
+        g_slice_free (PathData, path_data);
 }
 
 /**
@@ -686,7 +824,7 @@ unregister_hosting_server_handler (SoupServer        *server,
  *
  * Start hosting @local_path at @server_path. Files with the path
  * @local_path.LOCALE (if they exist) will be served up when LOCALE is
- * specified in the request's Accept-Language header. 
+ * specified in the request's Accept-Language header.
  **/
 void
 gupnp_context_host_path (GUPnPContext *context,
@@ -694,17 +832,21 @@ gupnp_context_host_path (GUPnPContext *context,
                          const char   *server_path)
 {
         SoupServer *server;
+        PathData *path_data;
 
         g_return_if_fail (GUPNP_IS_CONTEXT (context));
         g_return_if_fail (local_path != NULL);
         g_return_if_fail (server_path != NULL);
 
-        server = _gupnp_context_get_server (context);
+        server = gupnp_context_get_server (context);
 
-        soup_server_add_handler (server, server_path, NULL,
+        path_data = path_data_new (local_path, server_path);
+
+        soup_server_add_handler (server,
+                                 server_path,
                                  hosting_server_handler,
-                                 unregister_hosting_server_handler,
-                                 g_strdup (local_path));
+                                 path_data,
+                                 (GDestroyNotify) path_data_free);
 }
 
 /**
@@ -723,7 +865,7 @@ gupnp_context_unhost_path (GUPnPContext *context,
         g_return_if_fail (GUPNP_IS_CONTEXT (context));
         g_return_if_fail (server_path != NULL);
 
-        server = _gupnp_context_get_server (context);
+        server = gupnp_context_get_server (context);
 
         soup_server_remove_handler (server, server_path);
 }
